######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Provides an optional widget for Tool Specification Editor for each Tool Spec type (julia, python, executable, gams).

"""

from PySide6.QtCore import Signal, Slot
from PySide6.QtWidgets import QWidget
from PySide6.QtGui import QStandardItemModel, QStandardItem
from spine_engine.utils.helpers import resolve_conda_executable, resolve_python_interpreter
from spine_engine.execution_managers.conda_kernel_spec_manager import CondaKernelSpecManager
from spinetoolbox.helpers import busy_effect, file_is_valid, select_python_interpreter
from spinetoolbox.widgets.notification import Notification
from spinetoolbox.widgets.kernel_editor import KernelEditor, find_python_kernels
from spinetoolbox.qthread_pool_executor import QtBasedThreadPoolExecutor


class OptionalWidget(QWidget):
    def __init__(self, parent):
        super().__init__()
        self._parent = parent

    @property
    def _toolbox(self):
        return self._parent._toolbox

    @property
    def _project(self):
        return self._toolbox._project

    @property
    def _settings(self):
        return self._project._settings

    @property
    def _logger(self):
        return self._toolbox._logger

    def init_widget(self, specification):
        raise NotImplementedError

    def add_execution_settings(self):
        raise NotImplementedError


class PythonToolSpecOptionalWidget(OptionalWidget):
    _kernel_spec_data_ready = Signal(list, list)
    _kernel_spec_model_ready = Signal()

    def __init__(self, parent):
        """Init class."""
        from ..ui.python_kernel_spec_options import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(parent)
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        self.kernel_spec_model = QStandardItemModel(self)
        self.ui.comboBox_kernel_specs.setModel(self.kernel_spec_model)
        self._kernel_spec_editor = None
        self._kernel_spec_model_initialized = False
        self._saved_kernel_spec_name = None
        self._executor = QtBasedThreadPoolExecutor(max_workers=1)
        # Initialize UI elements with defaults
        use_jupyter_console = bool(
            int(self._toolbox.qsettings().value("appSettings/usePythonKernel", defaultValue="0"))
        )
        if use_jupyter_console:
            self.ui.radioButton_jupyter_console.setChecked(True)
            # Get the default kernel spec from Settings->Tools for new Jupyter Console Tool Specs
            self._saved_kernel_spec_name = self._toolbox.qsettings().value("appSettings/pythonKernel", defaultValue="")
        else:
            self.ui.radioButton_python_console.setChecked(True)
        default_python_path = self._toolbox.qsettings().value("appSettings/pythonPath", defaultValue="")
        self.ui.lineEdit_python_path.setPlaceholderText(resolve_python_interpreter(""))
        self.ui.lineEdit_python_path.setText(default_python_path)
        self.set_ui_for_jupyter_console(not use_jupyter_console)
        self.connect_signals()

    def connect_signals(self):
        """Connects signals."""
        self.ui.toolButton_refresh_kernel_specs.clicked.connect(self._refresh_kernel_spec_model)
        self.ui.comboBox_kernel_specs.activated.connect(self._parent._push_change_kernel_spec_command)
        self.ui.radioButton_jupyter_console.toggled.connect(self._parent._push_set_jupyter_console_mode)
        self.ui.toolButton_browse_python.clicked.connect(self.browse_python_button_clicked)
        self.ui.pushButton_open_kernel_spec_viewer.clicked.connect(self.show_python_kernel_spec_editor)
        self.ui.lineEdit_python_path.editingFinished.connect(self._parent._push_change_executable)
        self._kernel_spec_data_ready.connect(self._do_refresh_kernel_spec_model)
        self._kernel_spec_model_ready.connect(self._set_saved_kernel_spec)
        qApp.aboutToQuit.connect(self._executor.shutdown)  # pylint: disable=undefined-variable

    def init_widget(self, specification):
        """Initializes UI elements based on specification

        Args:
            specification (ToolSpecification): Specification to load
        """
        use_jupyter_console = specification.execution_settings["use_jupyter_console"]
        self.ui.radioButton_jupyter_console.blockSignals(True)
        self.ui.radioButton_python_console.blockSignals(True)
        if use_jupyter_console:
            self.ui.radioButton_jupyter_console.setChecked(True)
        else:
            self.ui.radioButton_python_console.setChecked(True)
        self.ui.radioButton_jupyter_console.blockSignals(False)
        self.ui.radioButton_python_console.blockSignals(False)
        self.set_ui_for_jupyter_console(not use_jupyter_console)
        # Must wait until model is initialized to set kernel spec name
        self._saved_kernel_spec_name = specification.execution_settings["kernel_spec_name"]
        self.set_executable(specification.execution_settings["executable"])

    @Slot()
    def _set_saved_kernel_spec(self):
        """Sets index of the kernel spec combobox to show the item that was saved with the Tool Specification.
        Make sure this is called after available kernel specs have been loaded."""
        if not self._saved_kernel_spec_name:
            self.ui.comboBox_kernel_specs.setCurrentIndex(0)  # Set 'Select kernel spec...'
        else:
            row = self.find_index_by_data(self._saved_kernel_spec_name)
            if row == -1:
                notification = Notification(
                    self._parent,
                    f"This Tool spec has kernel spec '{self._saved_kernel_spec_name}' "
                    f"saved but it could not be found.",
                )
                notification.show()
                row += 1  # Set 'Select kernel spec...'
            self.ui.comboBox_kernel_specs.setCurrentIndex(row)

    def add_execution_settings(self):
        """Collects execution settings based on optional widget state into a dictionary, which is returned."""
        idx = self.ui.comboBox_kernel_specs.currentIndex()
        if idx < 1:
            d = {"kernel_spec_name": "", "env": ""}
        else:
            item = self.ui.comboBox_kernel_specs.model().item(idx)
            k_spec_data = item.data()
            d = k_spec_data
        d["use_jupyter_console"] = self.ui.radioButton_jupyter_console.isChecked()
        self.validate_executable()  # Raises NameError if Python path is not valid
        d["executable"] = self.get_executable()
        return d

    @Slot(bool)
    def browse_python_button_clicked(self, _=False):
        """Calls static method that shows a file browser for selecting a Python interpreter."""
        select_python_interpreter(self, self.ui.lineEdit_python_path)
        self._parent._push_change_executable()

    def set_ui_for_jupyter_console(self, use_basic_console):
        """Enables or disables some UI elements in the optional widget according to checkBox state."""
        self.ui.lineEdit_python_path.setEnabled(use_basic_console)
        self.ui.toolButton_browse_python.setEnabled(use_basic_console)
        self.ui.comboBox_kernel_specs.setDisabled(use_basic_console)
        self.ui.toolButton_refresh_kernel_specs.setDisabled(use_basic_console)
        self.ui.pushButton_open_kernel_spec_viewer.setDisabled(use_basic_console)
        if not use_basic_console and not self._kernel_spec_model_initialized:
            self._refresh_kernel_spec_model()

    def validate_executable(self):
        """Check that Python path in the line edit is a file it exists and the file name starts with 'python'.

        Raises:
            NameError: If the python path in the line edit is not valid
        """
        p = self.ui.lineEdit_python_path.text().strip()
        if not file_is_valid(self._parent, p, "Invalid Python Interpreter", extra_check="python"):
            raise NameError

    def set_executable(self, p):
        self.ui.lineEdit_python_path.setText(p)

    def get_executable(self):
        return self.ui.lineEdit_python_path.text().strip()

    @Slot(bool)
    def show_python_kernel_spec_editor(self, _=False):
        """Opens kernel editor, where user can make kernel specs for the Jupyter Console."""
        p = self.ui.lineEdit_python_path.text()  # This may be an empty string
        j = ""
        item = self.kernel_spec_model.item(self.ui.comboBox_kernel_specs.currentIndex())
        if not item.data():
            selected_kernel_spec = ""
        else:
            selected_kernel_spec = item.data()["kernel_spec_name"]
        self._kernel_spec_editor = KernelEditor(self._parent, p, j, "python", selected_kernel_spec)
        self._kernel_spec_editor.finished.connect(self.python_kernel_editor_closed)
        self._kernel_spec_editor.open()

    @Slot(int)
    def python_kernel_editor_closed(self, ret_code):
        """Catches the selected Python kernel name when the editor is closed."""
        item = self.kernel_spec_model.item(self.ui.comboBox_kernel_specs.currentIndex())
        if not item.data():
            previous_kernel_spec = ""
        else:
            previous_kernel_spec = item.data()["kernel_spec_name"]
        self._refresh_kernel_spec_model()
        if ret_code != 1:  # Editor closed with Cancel
            # Set previous kernel selected in Python kernel combobox if it still exists
            r = self.find_index_by_data(previous_kernel_spec)
            if r == -1:
                self.ui.comboBox_kernel_specs.setCurrentIndex(0)  # Previous not found
            else:
                self.ui.comboBox_kernel_specs.setCurrentIndex(r)
            return
        new_kernel_spec = self._kernel_spec_editor.selected_kernel
        row = self.find_index_by_data(new_kernel_spec)
        if row == -1:  # New kernel spec not found, should be quite uncommon
            notification = Notification(self, f"Python kernel spec {new_kernel_spec} not found")
            notification.show()
            self.ui.comboBox_kernel_specs.setCurrentIndex(0)
        else:
            self._parent._push_change_kernel_spec_command(row)

    def find_index_by_data(self, string):
        """Searches the kernel spec model for the first item whose data matches the given string.
        Returns the items row number or -1 if not found."""
        if string == "":
            return 0
        for row in range(1, self.kernel_spec_model.rowCount()):  # Start from index 1
            index = self.kernel_spec_model.index(row, 0)
            item = self.kernel_spec_model.itemFromIndex(index)
            d = item.data()
            if d["kernel_spec_name"] == string:
                return row
        return -1

    @busy_effect
    def _get_all_kernel_specs(self):
        kernel_spec_data = []
        conda_kernel_spec_data = []
        kernel_specs = find_python_kernels()
        for n in kernel_specs:
            spec_data = {"kernel_spec_name": n, "env": ""}
            kernel_spec_data.append(spec_data)
        # Add auto-generated conda kernel spec names
        conda_exe = self._toolbox.qsettings().value("appSettings/condaPath", defaultValue="")
        conda_exe = resolve_conda_executable(conda_exe)
        if conda_exe != "":
            ksm = CondaKernelSpecManager(conda_exe=conda_exe)  # This is expensive
            conda_specs = ksm._all_specs()
            for i in conda_specs:
                spec_data = {"kernel_spec_name": i, "env": "conda"}
                conda_kernel_spec_data.append(spec_data)
        self._kernel_spec_data_ready.emit(kernel_spec_data, conda_kernel_spec_data)

    @Slot(bool)
    def _refresh_kernel_spec_model(self, _checked=False):
        self._kernel_spec_model_initialized = True
        self._executor.submit(self._get_all_kernel_specs)

    @Slot(list, list)
    def _do_refresh_kernel_spec_model(self, kernel_spec_data, conda_kernel_spec_data):
        item = self.kernel_spec_model.item(self.ui.comboBox_kernel_specs.currentIndex())
        if not item or not item.data():
            selected_kernel_spec = None
        else:
            selected_kernel_spec = item.data()["kernel_spec_name"]  # Remember the selected kernel spec
        self.kernel_spec_model.clear()
        first_item = QStandardItem("Select kernel spec...")
        self.kernel_spec_model.appendRow(first_item)
        # Add Python kernel specs
        for spec_data in kernel_spec_data:
            item = QStandardItem(spec_data["kernel_spec_name"] + " [Jupyter]")
            item.setData(spec_data)
            self.kernel_spec_model.appendRow(item)
        # Add auto-generated conda kernel spec names
        for spec_data in conda_kernel_spec_data:
            item = QStandardItem(spec_data["kernel_spec_name"] + " [Conda]")
            item.setData(spec_data)
            self.kernel_spec_model.appendRow(item)
        # Set the previously selected kernel spec as the current item after the model has been rebuilt
        if selected_kernel_spec is not None:
            row = self.find_index_by_data(selected_kernel_spec)
            if row == -1:
                # The kernel spec may have been removed
                self._parent._push_change_kernel_spec_command(0)
                return
            self.ui.comboBox_kernel_specs.setCurrentIndex(row)
        self._kernel_spec_model_ready.emit()


class ExecutableToolSpecOptionalWidget(OptionalWidget):
    def __init__(self, parent):
        """Init class."""
        from ..ui.executable_cmd_exec_options import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(parent)
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        self.shells = ["No shell", "cmd.exe", "powershell.exe", "bash"]
        self.ui.comboBox_shell.addItems(self.shells)
        self.connect_signals()

    def connect_signals(self):
        self.ui.lineEdit_command.editingFinished.connect(self._parent._push_change_executable_command)
        self.ui.comboBox_shell.activated.connect(self._parent._push_change_shell_command)

    def init_widget(self, specification):
        """Initializes UI elements based on specification."""
        self.ui.lineEdit_command.setText(specification.execution_settings["cmd"])
        shell = specification.execution_settings["shell"]
        ind = next(iter(k for k, t in enumerate(self.shells) if t.lower() == shell), 0)
        self.ui.comboBox_shell.setCurrentIndex(ind)

    def add_execution_settings(self):
        """Collects execution settings based on optional widget state into a dictionary, which is returned."""
        return {"cmd": self.ui.lineEdit_command.text(), "shell": self.get_current_shell()}

    def get_current_shell(self):
        """Returns the selected shell in the shell combo box."""
        ind = self.ui.comboBox_shell.currentIndex()
        if ind < 1:
            return ""
        return self.ui.comboBox_shell.currentText()



# -*- coding: utf-8 -*-
######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

################################################################################
## Form generated from reading UI file 'parameter_renamer_editor.ui'
##
## Created by: Qt User Interface Compiler version 6.4.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QHBoxLayout, QHeaderView,
    QPushButton, QSizePolicy, QSpacerItem, QSplitter,
    QTreeWidgetItem, QVBoxLayout, QWidget)

from ..widgets.parameter_tree_widget import ParameterTreeWidget
from spine_items.data_transformer.widgets.drop_target_table import DropTargetTable

class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName(u"Form")
        Form.resize(616, 530)
        self.remove_parameter_action = QAction(Form)
        self.remove_parameter_action.setObjectName(u"remove_parameter_action")
        self.remove_parameter_action.setShortcutContext(Qt.WidgetShortcut)
        self.verticalLayout_2 = QVBoxLayout(Form)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.splitter = QSplitter(Form)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Orientation.Horizontal)
        self.available_parameters_tree_view = ParameterTreeWidget(self.splitter)
        __qtreewidgetitem = QTreeWidgetItem()
        __qtreewidgetitem.setText(0, u"1");
        self.available_parameters_tree_view.setHeaderItem(__qtreewidgetitem)
        self.available_parameters_tree_view.setObjectName(u"available_parameters_tree_view")
        self.available_parameters_tree_view.setDragEnabled(True)
        self.available_parameters_tree_view.setDragDropMode(QAbstractItemView.DragOnly)
        self.available_parameters_tree_view.setDefaultDropAction(Qt.CopyAction)
        self.splitter.addWidget(self.available_parameters_tree_view)
        self.available_parameters_tree_view.header().setVisible(False)
        self.verticalLayoutWidget = QWidget(self.splitter)
        self.verticalLayoutWidget.setObjectName(u"verticalLayoutWidget")
        self.verticalLayout = QVBoxLayout(self.verticalLayoutWidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.add_parameter_button = QPushButton(self.verticalLayoutWidget)
        self.add_parameter_button.setObjectName(u"add_parameter_button")

        self.horizontalLayout.addWidget(self.add_parameter_button)

        self.remove_parameter_button = QPushButton(self.verticalLayoutWidget)
        self.remove_parameter_button.setObjectName(u"remove_parameter_button")

        self.horizontalLayout.addWidget(self.remove_parameter_button)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.renames_table_view = DropTargetTable(self.verticalLayoutWidget)
        self.renames_table_view.setObjectName(u"renames_table_view")
        self.renames_table_view.setDragDropMode(QAbstractItemView.DropOnly)

        self.verticalLayout.addWidget(self.renames_table_view)

        self.splitter.addWidget(self.verticalLayoutWidget)

        self.verticalLayout_2.addWidget(self.splitter)


        self.retranslateUi(Form)

        QMetaObject.connectSlotsByName(Form)
    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", u"Form", None))
        self.remove_parameter_action.setText(QCoreApplication.translate("Form", u"Remove parameter", None))
#if QT_CONFIG(shortcut)
        self.remove_parameter_action.setShortcut(QCoreApplication.translate("Form", u"Del", None))
#endif // QT_CONFIG(shortcut)
        self.add_parameter_button.setText(QCoreApplication.translate("Form", u"Add", None))
        self.remove_parameter_button.setText(QCoreApplication.translate("Form", u"Remove", None))
    # retranslateUi




######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Contains a model for Exporter's output preview.

"""
from PySide6.QtCore import QAbstractListModel, QModelIndex, Qt
from spine_items.utils import Database


class DatabaseListModel(QAbstractListModel):
    """A model for exporter database lists."""

    def __init__(self, databases):
        """
        Args:
            databases (list of Database): databases to list
        """
        super().__init__()
        self._databases = databases

    def add(self, database):
        """
        Appends a database to the list.

        Args:
            database (Database): a database to add
        """
        row = len(self._databases)
        self.beginInsertRows(QModelIndex(), row, row)
        self._databases.append(database)
        self.endInsertRows()

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if not index.isValid():
            return None
        if role == Qt.ItemDataRole.DisplayRole:
            return self._databases[index.row()].url
        return None

    def insertRows(self, row, count, parent=QModelIndex()):
        self.beginInsertRows(parent, row, row + count - 1)
        self._databases = self._databases[:row] + [Database() for _ in range(count)] + self._databases[row:]
        self.endInsertRows()

    def item(self, url):
        """
        Returns database item for given URL.

        Args:
            url (str): database URL

        Returns:
            Database: a database
        """
        for db in self._databases:
            if db.url == url:
                return db
        raise RuntimeError(f"Database '{url}' not found.")

    def items(self):
        """
        Returns a list of databases this model contains.

        Returns:
            list of Database: database
        """
        return self._databases

    def remove(self, url):
        """
        Removes database item with given URL.

        Args:
            url (str): database URL

        Returns:
            Database: removed database or None if not found
        """
        for row, db in enumerate(self._databases):
            if db.url == url:
                self.removeRows(row, 1)
                return db
        return None

    def removeRows(self, row, count, parent=QModelIndex()):
        self.beginRemoveRows(parent, row, row + count - 1)
        self._databases = self._databases[:row] + self._databases[row + count :]
        self.endRemoveRows()

    def rowCount(self, parent=QModelIndex()):
        return len(self._databases)

    def update_url(self, old, new):
        """
        Updates a database URL.

        Args:
            old (str): old URL
            new (str): new URL
        """
        for row, db in enumerate(self._databases):
            if old == db.url:
                db.url = new
                index = self.index(row, 0)
                self.dataChanged.emit(index, index, [Qt.ItemDataRole.DisplayRole])
                return

    def urls(self):
        """
        Returns database URLs.

        Returns:
            set of str: database URLs
        """
        return {db.url for db in self._databases}



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################
"""
Contains :class:`PreviewUpdater`.

"""
from copy import deepcopy
from time import monotonic
from PySide6.QtCore import QItemSelectionModel, QModelIndex, QObject, QRunnable, Qt, QThreadPool, Signal, Slot
from PySide6.QtWidgets import QFileDialog

from spinedb_api.export_mapping.group_functions import NoGroup
from spinedb_api.spine_io.exporters.writer import write
from spinedb_api import DatabaseMapping, SpineDBVersionError, SpineDBAPIError
from spinetoolbox.helpers import busy_effect
from ..mvcmodels.full_url_list_model import FullUrlListModel
from ..mvcmodels.mappings_table_model import MappingsTableModel
from ..mvcmodels.preview_tree_model import PreviewTreeModel
from ..mvcmodels.preview_table_model import PreviewTableModel
from ..preview_table_writer import TableWriter


class PreviewUpdater:
    def __init__(
        self, window, ui, url_model, mappings_table_model, mappings_proxy_model, mapping_editor_table_model, project_dir
    ):
        """
        Args:
            window (SpecificationEditorWindow): specification editor's window
            ui (Ui_Form): specification editor's UI
            url_model (FullUrlListModel, optional): URL model
            mappings_table_model (MappingsTableModel): mappings table model
            mappings_proxy_model (QAbstractProxyModel): proxy model for mappings table
            mapping_editor_table_model (MappingEditorTableModel): mapping table model
            project_dir (str): path to initial directory from which to load databases
        """
        self._current_url = None
        self._window = window
        self._ui = ui
        self._project_dir = project_dir
        if url_model is None:
            url_model = FullUrlListModel(self._ui.database_url_combo_box)
        self._url_model = url_model
        self._url_model.rowsInserted.connect(self._enable_controls_after_url_insertion)
        self._url_model.modelReset.connect(self._enable_controls)
        self._url_model.destroyed.connect(self._forget_url_model)
        self._ui.database_url_combo_box.setModel(self._url_model)
        self._mappings_table_model = mappings_table_model
        self._set_expect_removals_and_inserts(False)
        self._mappings_table_model.write_order_about_to_change.connect(
            lambda: self._set_expect_removals_and_inserts(True)
        )
        self._mappings_table_model.write_order_changed.connect(lambda: self._set_expect_removals_and_inserts(False))
        self._mappings_table_model.dataChanged.connect(self._rename_mapping)
        self._mappings_table_model.dataChanged.connect(self._enable_mapping)
        self._mappings_table_model.dataChanged.connect(self._update_changed_tables)
        self._mappings_proxy_model = mappings_proxy_model
        self._mapping_editor_table_model = mapping_editor_table_model
        self._mapping_editor_table_model.dataChanged.connect(lambda *_: self._update_current_mappings_tables())
        self._expect_mapping_editor_resets(False)
        self._window.current_mapping_about_to_change.connect(lambda: self._expect_mapping_editor_resets(True))
        self._window.current_mapping_changed.connect(lambda: self._expect_mapping_editor_resets(False))
        self._preview_tree_model = PreviewTreeModel()
        self._preview_tree_model.dataChanged.connect(self._update_table)
        self._preview_tree_model.dataChanged.connect(self._expand_tree_after_table_change)
        self._preview_tree_model.rowsInserted.connect(self._expand_tree_after_table_insert)
        self._preview_table_model = PreviewTableModel()
        self._stamps = dict()
        self._thread_pool = QThreadPool()
        self._mapping_tables = dict()
        self._ui.preview_tree_view.setModel(self._preview_tree_model)
        self._ui.preview_tree_view.selectionModel().currentChanged.connect(self._change_table)
        self._ui.preview_table_view.setModel(self._preview_table_model)
        self._ui.database_url_combo_box.currentTextChanged.connect(self._reload_preview)
        self._ui.live_preview_check_box.clicked.connect(self._reload_preview)
        self._ui.max_preview_rows_spin_box.valueChanged.connect(self._reload_preview)
        self._ui.max_preview_tables_spin_box.valueChanged.connect(self._reload_preview)
        self._ui.load_url_from_fs_button.clicked.connect(self._load_url_from_filesystem)
        self._ui.mappings_table.selectionModel().currentChanged.connect(self._change_selected_table)
        self._enable_controls()
        if self._url_model.rowCount() > 0:
            self._reload_preview()

    @Slot()
    def _reload_preview(self):
        """Sets the current url and reloads preview."""
        current_url = self._ui.database_url_combo_box.currentText()
        if not current_url:
            return
        self._current_url = current_url
        self._stamps.clear()
        for row in range(self._mappings_table_model.rowCount()):
            index = self._mappings_table_model.index(row, 0)
            if index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Checked.value:
                self._load_preview_data(index.data())

    @Slot(QModelIndex, QModelIndex)
    def _change_table(self, current, previous):
        """
        Changes preview table data.

        Args:
            current (QModelIndex): index to the currently selected table on the list
            previous (QModelIndex): index to the previously selected table on the list
        """
        current_parent = self._preview_tree_model.parent(current)
        if not current_parent.isValid():
            current_parent = current
        previous_parent = self._preview_tree_model.parent(previous)
        if not previous_parent.isValid():
            previous_parent = previous
        if current_parent != previous_parent:
            name = self._preview_tree_model.data(current_parent)
            index = self._mappings_proxy_model.mapFromSource(self._mappings_table_model.index_of(name))
            self._ui.mappings_table.selectionModel().currentChanged.disconnect(self._change_selected_table)
            self._ui.mappings_table.setCurrentIndex(index)
            self._ui.mappings_table.selectionModel().currentChanged.connect(self._change_selected_table)
        table = current.data(PreviewTreeModel.TABLE_ROLE)
        if table is not None:
            table_name = current.data()
            mapping_name = current.parent().data()
            mapping_colors = self._mapping_editor_table_model.mapping_colors()
            self._preview_table_model.reset(mapping_name, table_name, table, mapping_colors)
        else:
            self._preview_table_model.clear()

    @Slot(QModelIndex, QModelIndex)
    def _change_selected_table(self, current, previous):
        """
        Changes selected table on the tree view after user has selected another mapping.

        Args:
            current (QModelIndex): index to selected mapping on mapping list
            previous (QModelIndex): index to previously selected mapping on mapping list
        """
        if not current.isValid():
            return
        current = self._mappings_proxy_model.mapToSource(current)
        mappings_index = self._mappings_table_model.index(current.row(), 0)
        if mappings_index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Unchecked.value:
            return
        current_preview = self._ui.preview_tree_view.selectionModel().currentIndex()
        if current_preview.parent().row() >= self._preview_tree_model.rowCount():
            # May happen when mappings at the end of the list are removed.
            return
        mapping_name = mappings_index.data()
        if current_preview.parent().data() == mapping_name:
            return
        self._set_current_table(mapping_name)

    def _set_current_table(self, name):
        """Sets the current index of preview tree.

        Args:
            name (str): mapping's name
        """
        for row in range(self._preview_tree_model.rowCount()):
            index = self._preview_tree_model.index(row, 0)
            if index.data() == name:
                if self._preview_tree_model.rowCount(index) > 0:
                    table_index = self._preview_tree_model.index(0, 0, index)
                    self._ui.preview_tree_view.selectionModel().setCurrentIndex(
                        table_index, QItemSelectionModel.ClearAndSelect
                    )
                    return
                self._ui.preview_tree_view.selectionModel().setCurrentIndex(index, QItemSelectionModel.ClearAndSelect)
                return

    @Slot(QModelIndex, QModelIndex, list)
    def _expand_tree_after_table_change(self, top_left, bottom_right, roles):
        """
        Expands preview tree's root items if their tables have received updates.

        Args:
            top_left (QModelIndex): top left corner of changed preview tree data
            bottom_right (QModelIndex): bottom right corner of changed preview tree data
            roles (list of int): roles
        """
        if PreviewTreeModel.TABLE_ROLE not in roles:
            return
        root_index = top_left.parent()
        if not root_index.isValid():
            return
        self._ui.preview_tree_view.setExpanded(root_index, True)

    @Slot(QModelIndex, QModelIndex, list)
    def _expand_tree_after_table_insert(self, parent, first, last):
        """
        Expands preview tree's root items if tables have been added/removed.

        Args:
            parent (QModelIndex): parent index of items that have been added/removed
            first (int): first row index of affected items
            last (int): last row index of affected items
        """
        if not parent.isValid():
            return
        self._ui.preview_tree_view.setExpanded(parent, True)

    @Slot(QModelIndex, QModelIndex, list)
    def _update_table(self, top_left, bottom_right, roles):
        """
        Updates the preview table if its data has changed.

        Args:
            top_left (QModelIndex): top left corner of changed preview tree data
            bottom_right (QModelIndex): bottom right corner of changed preview tree data
            roles (list of int): roles
        """
        if PreviewTreeModel.TABLE_ROLE not in roles:
            return
        current_parent = top_left.parent()
        current_table_name = self._preview_table_model.table_name()
        current_mapping_name = self._preview_table_model.mapping_name()
        for row in range(top_left.row(), bottom_right.row() + 1):
            index = self._preview_tree_model.index(row, 0, current_parent)
            table_name = index.data()
            mapping_name = index.parent().data()
            if current_table_name != table_name or current_mapping_name != mapping_name:
                continue
            table = index.data(PreviewTreeModel.TABLE_ROLE)
            mapping_colors = self._mapping_editor_table_model.mapping_colors()
            self._preview_table_model.reset(mapping_name, table_name, table, mapping_colors)
            break

    def _set_expect_removals_and_inserts(self, except_removals):
        """Connects and disconnects signals with regarding mapping removals.

        Args:
            except_removals (bool): True means that a mapping removal is expected
        """
        if except_removals:
            self._mappings_table_model.rowsAboutToBeRemoved.disconnect(self._remove_mappings)
            self._mappings_table_model.rowsInserted.disconnect(self._add_mappings)
        else:
            self._mappings_table_model.rowsAboutToBeRemoved.connect(self._remove_mappings)
            self._mappings_table_model.rowsInserted.connect(self._add_mappings)

    @Slot(int, int, QModelIndex)
    def _remove_mappings(self, parent, first, last):
        """
        Removes mappings from the preview models.

        Args:
            parent (QModelIndex): parent index, ignored
            first (int): first mapping list row to be removed
            last (int): last mapping list row to be removed
        """
        if self._current_url is None:
            return
        for row in range(last, first - 1, -1):
            index = self._mappings_table_model.index(row, 0)
            if index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Unchecked.value:
                continue
            removed_name = index.data()
            self._preview_tree_model.remove_mapping(removed_name)
            self._stamps.pop((self._current_url, removed_name), None)

    @Slot(QModelIndex, QModelIndex, list)
    def _rename_mapping(self, top_left, bottom_right, roles):
        """
        Renames preview models' mappings.

        Args:
            top_left (QModelIndex): top left corner of modified mappings' in mapping list model
            bottom_right (QModelIndex): bottom right corner of modified mappings' in mapping list model
            roles (list of int): changed data's role
        """
        if Qt.ItemDataRole.DisplayRole not in roles or self._current_url is None:
            return
        make_index = self._mappings_table_model.index
        indexes = [make_index(row, 0) for row in range(self._mappings_table_model.rowCount())]
        names = [
            index.data()
            for index in indexes
            if index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Checked.value
        ]
        old_name, new_name = self._preview_tree_model.rename_mappings(names)
        if not old_name:
            return
        self._stamps.pop((self._current_url, old_name), None)
        self._set_current_table(new_name)

    @Slot(QModelIndex, QModelIndex, list)
    def _enable_mapping(self, top_left, bottom_right, roles):
        """
        Removes/adds mappings as they are enabled or disabled.

        Args:
            top_left (QModelIndex): top left corner of modified mappings' in mapping list model
            bottom_right (QModelIndex): bottom right corner of modified mappings' in mapping list model
            roles (list of int): changed data's role
        """
        if Qt.ItemDataRole.CheckStateRole.value not in roles or self._current_url is None:
            return
        first = top_left.row()
        last = bottom_right.row()
        for row in range(first, last + 1):
            index = self._mappings_table_model.index(row, 0)
            enabled = index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Checked.value
            name = index.data()
            if not enabled and self._preview_tree_model.has_name(name):
                self._preview_tree_model.remove_mapping(name)
                self._stamps.pop((self._current_url, name), None)
            elif not self._preview_tree_model.has_name(name):
                self._load_preview_data(name)

    @Slot(QModelIndex, QModelIndex, list)
    def _update_changed_tables(self, top_left, bottom_right, roles):
        """
        Updates mappings whenever the always write header flag or fixed table name changes.

        Args:
            top_left (QModelIndex): top left corner of modified mappings' in mapping list model
            bottom_right (QModelIndex): bottom right corner of modified mappings' in mapping list model
            roles (list of int): changed data's role
        """
        if (
            not {
                MappingsTableModel.ALWAYS_EXPORT_HEADER_ROLE,
                MappingsTableModel.FIXED_TABLE_NAME_ROLE,
                MappingsTableModel.GROUP_FN_ROLE,
            }
            & set(roles)
        ):
            return
        row = self._mappings_proxy_model.mapToSource(self._ui.mappings_table.currentIndex()).row()
        index = self._mappings_table_model.index(row, 0)
        if index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Unchecked.value:
            return
        name = index.data()
        self._load_preview_data(name)

    @Slot()
    def _update_current_mappings_tables(self):
        """Reloads mapping that is currently selected on mapping list."""
        if self._current_url is None:
            return
        current_index = self._ui.mappings_table.currentIndex()
        if not current_index.isValid():
            return
        row = self._mappings_proxy_model.mapToSource(current_index).row()
        index = self._mappings_table_model.index(row, 0)
        if index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Unchecked.value:
            return
        mapping_name = index.data()
        self._load_preview_data(mapping_name)

    @Slot(QModelIndex, int, int)
    def _add_mappings(self, parent, first, last):
        """Adds new mapping for preview.

        Args:
            parent (QModelIndex): parent index - unused
            first (int): first row added to mappings table
            last (int): last row added to mappings table
        """
        if self._current_url is None:
            return
        for row in range(first, last + 1):
            index = self._mappings_table_model.index(first, 0)
            if index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Unchecked.value:
                continue
            mapping_name = index.data()
            self._load_preview_data(mapping_name)

    def _expect_mapping_editor_resets(self, expect_resets):
        """Connects and disconnects signals that handle mapping editor resets.

        Args:
            expect_resets (bool): True to except editor resets
        """
        if expect_resets:
            self._mapping_editor_table_model.modelReset.disconnect(self._update_current_mappings_tables)
        else:
            self._mapping_editor_table_model.modelReset.connect(self._update_current_mappings_tables)

    @Slot(QModelIndex, int, int)
    def _enable_controls_after_url_insertion(self, parent, first, last):
        """Enables controls after new preview database URL has been inserted."""
        self._enable_controls()

    @Slot()
    def _enable_controls(self):
        """Enables and disables widgets as needed."""
        urls_available = self._url_model.rowCount() > 0
        self._ui.live_preview_check_box.setEnabled(urls_available)
        self._ui.max_preview_rows_spin_box.setEnabled(urls_available)
        self._ui.max_preview_tables_spin_box.setEnabled(urls_available)

    def _load_preview_data(self, mapping_name):
        """
        Loads preview data from database into the preview tables if the url is set and live previews are enabled.

        Args:
            mapping_name (str): mapping's name
        """
        if self._current_url is None or not self._ui.live_preview_check_box.isChecked() or not mapping_name:
            return
        mapping_spec = self._mappings_table_model.mapping_specification(mapping_name)
        max_tables = self._ui.max_preview_tables_spin_box.value()
        max_rows = self._ui.max_preview_rows_spin_box.value()
        id_ = (self._current_url, mapping_name)
        stamp = monotonic()
        self._stamps[id_] = stamp
        worker = _Worker(
            self._current_url,
            mapping_name,
            deepcopy(mapping_spec.root),
            mapping_spec.always_export_header,
            stamp,
            max_tables,
            max_rows,
            mapping_spec.group_fn,
        )
        worker.signals.table_written.connect(self._add_or_update_data)
        self._thread_pool.start(worker)

    @Slot(tuple, str, dict, float)
    def _add_or_update_data(self, worker_id, mapping_name, data, stamp):
        """
        Sets preview data.

        Args:
            worker_id (tuple): a worker identifier
            mapping_name (str): mapping's name
            data (dict): mapping from table name to table
            stamp (float): worker's time stamp
        """
        current_stamp = self._stamps.pop(worker_id, None)
        if stamp != current_stamp:
            if current_stamp is not None:
                self._stamps[worker_id] = current_stamp
            return
        self._preview_tree_model.add_or_update_tables(mapping_name, data)
        mapping_index = self._ui.mappings_table.selectionModel().currentIndex()
        self._change_selected_table(mapping_index, mapping_index)

    @Slot(bool)
    def _load_url_from_filesystem(self, _):
        """Shows a file dialog and opens selected database in the preview widget."""
        path = self._browse_database()
        if not path:
            return
        url = "sqlite:///" + path
        self._url_model.append(url)
        self._ui.database_url_combo_box.setCurrentText(url)

    def _browse_database(self):
        """
        Queries a database file from the user.

        Returns:
            str: path to database file
        """
        return QFileDialog.getOpenFileName(self._window, "Select database", self._project_dir, "sqlite (*.sqlite)")[0]

    @Slot()
    def _forget_url_model(self):
        """Replaces current URL model with a mock one."""
        self._url_model = FullUrlListModel(self._ui.database_url_combo_box)
        self._url_model.rowsInserted.connect(self._enable_controls_after_url_insertion)
        self._url_model.modelReset.connect(self._enable_controls)
        self._url_model.destroyed.connect(self._forget_url_model)
        self._ui.database_url_combo_box.setModel(self._url_model)
        self._reload_preview()

    def tear_down(self):
        """Stops all workers."""
        self._stamps.clear()
        self._thread_pool.clear()
        self._thread_pool.deleteLater()
        self._url_model.rowsInserted.disconnect(self._enable_controls_after_url_insertion)
        self._url_model.modelReset.disconnect(self._enable_controls)
        self._url_model.destroyed.disconnect(self._forget_url_model)


class _Worker(QRunnable):
    class Signals(QObject):
        table_written = Signal(tuple, str, dict, float)

    def __init__(
        self, url, mapping_name, mapping, always_export_header, stamp, max_tables=20, max_rows=20, group_fn=NoGroup.NAME
    ):
        super().__init__()
        self._url = url
        self._mapping_name = mapping_name
        self._mapping = mapping
        self._always_export_header = always_export_header
        self._max_tables = max_tables
        self._max_rows = max_rows
        self._group_fn = group_fn
        self._stamp = stamp
        self.signals = self.Signals()

    @busy_effect
    def run(self):
        try:
            db_map = DatabaseMapping(self._url)
        except SpineDBVersionError:
            tables = {"error": [["unsupported database version"]]}
            self.signals.table_written.emit((self._url, self._mapping_name), self._mapping_name, tables, self._stamp)
            return
        except SpineDBAPIError as error:
            tables = {"error": [[str(error)]]}
            self.signals.table_written.emit((self._url, self._mapping_name), self._mapping_name, tables, self._stamp)
            return
        try:
            writer = TableWriter()
            write(
                db_map,
                writer,
                self._mapping,
                empty_data_header=self._always_export_header,
                max_tables=self._max_tables,
                max_rows=self._max_rows,
                group_fns=self._group_fn,
            )
            self.signals.table_written.emit(
                (self._url, self._mapping_name), self._mapping_name, writer.tables, self._stamp
            )
        except SpineDBAPIError as error:
            tables = {"error": [[str(error)]]}
            self.signals.table_written.emit((self._url, self._mapping_name), self._mapping_name, tables, self._stamp)
            return
        finally:
            db_map.connection.close()
            self.signals.deleteLater()
